@@ -1,4 +1,4 @@
-/* Copyright (c) 2002,2003 Sam Trenholme
+/* Copyright (c) 2002-2007 Sam Trenholme
  *
  * TERMS
  *
@@ -16,8 +16,6 @@
  * fitness for purpose.
  */
 
-/* This is the core DNS server */
-
 /* Langauge-specific labels */
 #include "MaraBigHash_locale.h"
 
@@ -32,17 +30,27 @@
 #include <string.h>
 #include <time.h>
 #include <sys/types.h>
+#ifndef MINGW32
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#else
+#include <winsock.h>
+#include <wininet.h>
+#endif
 #include "../dns/functions_dns.h"
 #include "../parse/functions_parse.h"
+#include "../parse/Csv2_database.h"
+#include "../parse/Csv2_read.h"
+#include "../parse/Csv2_functions.h"
 #include "functions_server.h"
+#include "timestamp.h"
 
 extern int log_level;
+extern int no_cname_warnings;
+extern int dns_records_served;
+int cname_warnings_shown = 0;
 
-extern int mhash_firstkey();
-extern int mhash_nextkey();
 
 /* Handler to handle warnings.
    Input: Pointer to js_string object with warning message,
@@ -173,6 +181,9 @@ void init_rr(rr *data) {
      data->seen = 0;
      data->zap = 0;
      data->expire = data->ttl = data->authoritative = 0;
+     data->perms = 0;
+     data->list = 0;
+     data->rcode = 0;
      return;
      }
 
@@ -183,10 +194,10 @@ void init_rr(rr *data) {
 */
 
 int mhash_put_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
-                 uint32 authoritative, uint32 expire) {
+                 uint32 authoritative, uint32 expire, perm_t perms) {
 
-    rr *data;
-    js_string *new;
+    rr *data = 0;
+    js_string *new = 0;
     int ret, rrtype;
     /* Create a structure for putting the rr data in */
     if((data = js_alloc(1,sizeof(rr))) == 0)
@@ -199,6 +210,7 @@ int mhash_put_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
     data->expire = 0;
     data->ttl = ttl;
     data->authoritative = authoritative;
+    data->perms = perms;
     /* Get the rr type from the query string */
     rrtype = get_rtype(query);
     if(rrtype == JS_ERROR) {
@@ -230,6 +242,11 @@ int mhash_put_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
     /* Store that data in the big hash */
     ret = mhash_put(hash,query,data,MARA_DNSRR);
 
+    /* Add this entry to the corresponding ANY chain */
+    if(any_add_rr(hash,query,data) == JS_ERROR) {
+        return JS_ERROR;
+        }
+
     /* Automatically grow the hash if appropriate */
     mhash_autogrow(hash);
 
@@ -257,7 +274,7 @@ int mhash_put_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
 */
 
 int mhash_add_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
-                 uint32 authoritative, uint32 expire) {
+                 uint32 authoritative, uint32 expire, perm_t perms) {
 
     rr *data, *point;
     js_string *new;
@@ -275,6 +292,7 @@ int mhash_add_rr(mhash *hash, js_string *query, js_string *value, uint32 ttl,
     data->expire = 0;
     data->ttl = ttl;
     data->authoritative = authoritative;
+    data->perms = perms;
 
     /* Get the rr type from the query string */
     rrtype = get_rtype(query);
@@ -407,8 +425,8 @@ int mhash_add_ptr(mhash *hash, js_string *query, js_string *value) {
 
 int mhash_add_ip(mhash *hash, js_string *query, js_string *value) {
 
-    rr *data, *point;
-    js_string *new, *new_query;
+    rr *data = 0, *point = 0;
+    js_string *new = 0, *new_query = 0;
     mhash_e spot_data;
     int rrtype;
 
@@ -560,7 +578,8 @@ int mhash_add_ns(mhash *hash, js_string *query, js_string *ns) {
     /* Next: Find the NS RR (or RR set) we will point to */
     spot_data = mhash_get(hash,ns);
     if(spot_data.value == 0) { /* If the NS element does not exist */
-        return JS_ERROR;
+        return JS_SUCCESS; /* We may not have a NS record in place yet;
+                              just make it an RR without a NS then */
         }
     /* Exit if the NS data pointed to is *not* a DNS resource record */
     if(spot_data.datatype != MARA_DNSRR) {
@@ -582,7 +601,7 @@ int mhash_add_ns(mhash *hash, js_string *query, js_string *ns) {
 
 /* Populate the main assosciative array (the one where the raw UDP query is
    the key and the value is the answer) with the data from the various
-   csv1 files.
+   csv1 files (this is called from populate_main)
    Input: A pointer to the hash to populate, a pointer to the string to
           put an error message in, whether MaraDNS is being recursive or
           not
@@ -591,7 +610,7 @@ int mhash_add_ns(mhash *hash, js_string *query, js_string *ns) {
    Global vars used: The kvars and dvars
  */
 
-int populate_main(mhash *maintable, js_string *error, int recursive) {
+int parse_csv1s(mhash *maintable, js_string *error, int recursive) {
     mhash *csvs; /* List of CSV1 files to read */
     js_string *zone;
     js_string *udpzone, *filename, *line, *soaline, *pline;
@@ -610,38 +629,102 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
 
     /* Create the js_String objects containing the zone name and
        corresponding filename */
-    if((udpzone = js_create(128,1)) == 0)
+    if((udpzone = js_create(128,1)) == 0) {
         return JS_ERROR;
-    if((filename = js_create(256,1)) == 0)
+        }
+    if((filename = js_create(256,1)) == 0) {
+        js_destroy(udpzone);
         return JS_ERROR;
-    if((line = js_create(MAX_RECORD_LENGTH + 3,1)) == 0)
+        }
+    if((line = js_create(MAX_RECORD_LENGTH + 3,1)) == 0) {
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if(js_set_encode(line,MARA_LOCALE) == JS_ERROR)
+        }
+    if(js_set_encode(line,MARA_LOCALE) == JS_ERROR) {
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if((pline = js_create(MAX_RECORD_LENGTH + 3,1)) == 0)
+        }
+    if((pline = js_create(MAX_RECORD_LENGTH + 3,1)) == 0) {
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if(js_set_encode(pline,MARA_LOCALE) == JS_ERROR)
+        }
+    if(js_set_encode(pline,MARA_LOCALE) == JS_ERROR) {
+        js_destroy(pline);
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if((soaline = js_create(256,1)) == 0)
+        }
+    if((soaline = js_create(256,1)) == 0) {
+        js_destroy(pline);
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if(js_set_encode(soaline,MARA_LOCALE) == JS_ERROR)
+        }
+    if(js_set_encode(soaline,MARA_LOCALE) == JS_ERROR) {
+        js_destroy(soaline);
+        js_destroy(pline);
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if((query = js_create(256,1)) == 0)
+        }
+    if((query = js_create(256,1)) == 0) {
+        js_destroy(soaline);
+        js_destroy(pline);
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if((data = js_create(MAX_RECORD_LENGTH + 3,1)) == 0)
+        }
+    if((data = js_create(MAX_RECORD_LENGTH + 3,1)) == 0) {
+        js_destroy(query);
+        js_destroy(soaline);
+        js_destroy(pline);
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if((zone = js_create(MAX_ZONE_SIZE,1)) == 0)
+        }
+    if((zone = js_create(MAX_ZONE_SIZE,1)) == 0) {
+        js_destroy(data);
+        js_destroy(query);
+        js_destroy(soaline);
+        js_destroy(pline);
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
-    if((desc = js_alloc(1,sizeof(js_file))) == 0)
+        }
+    if((desc = js_alloc(1,sizeof(js_file))) == 0) {
+        js_destroy(zone);
+        js_destroy(data);
+        js_destroy(query);
+        js_destroy(soaline);
+        js_destroy(pline);
+        js_destroy(line);
+        js_destroy(filename);
+        js_destroy(udpzone);
         return JS_ERROR;
+        }
 
     /* Pass 1: fill up the hash with the AN and NS answers */
 
     /* Get all of the keys in the csv1 hash */
-    csvs = (mhash *)dvar_raw(0);
+    csvs = (mhash *)dvar_raw(dq_keyword2n("csv1"));
+
     if(csvs == 0) {
         /* A csv1 hash is no longer mandatory, now that MaraDNS has recursive
            capability */
+        js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+        js_destroy(pline); js_destroy(soaline); js_destroy(query);
+        js_destroy(data); js_destroy(zone); js_dealloc(desc);
         return 0;
         /*js_qstr2js(error,L_INITCSV); */
         /*"csv1 hash is not correctly initialized.\nMake sure to have csv1 initialized with csv1 = {}, and that it has one or\nmore elements." */
@@ -651,6 +734,9 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
     if(mhash_firstkey(csvs,zone) == 0) {
         /* We actually do not need to have anything in the hash, in
            the case of people running a recursive nameserver */
+        js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+        js_destroy(pline); js_destroy(soaline); js_destroy(query);
+        js_destroy(data); js_destroy(zone); js_dealloc(desc);
         return 0;
         }
     do {
@@ -659,10 +745,18 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
         if(zone != 0) {
 
             /* Make udpzone be like zone, but in raw RFC1035 format */
-            if(js_qstr2js(udpzone,"A") == JS_ERROR)
+            if(js_qstr2js(udpzone,"A") == JS_ERROR) {
+                js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+                js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                js_destroy(data); js_destroy(zone); js_dealloc(desc);
                 return JS_ERROR;
-            if(js_append(zone,udpzone) == JS_ERROR)
+                }
+            if(js_append(zone,udpzone) == JS_ERROR) {
+                js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+                js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                js_destroy(data); js_destroy(zone); js_dealloc(desc);
                 return JS_ERROR;
+                }
             if(hname_2rfc1035(udpzone) == JS_ERROR) {
                 js_qstr2js(error,L_BADZONE); /* "A zone file is incorrectly named.  All zone files must end with a dot, e.g.\ncsv1[\"example.com.\"] = \"filename\".\nBad zone name: " */
                 js_qappend("\ncsv1[\"",error);
@@ -670,16 +764,28 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
                 js_qappend("\"] = \"",error);
                 js_append(mhash_get_js(csvs,zone),error);
                 js_qappend("\"\n",error);
+                js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+                js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                js_destroy(data); js_destroy(zone); js_dealloc(desc);
                 return -2;
                 }
             /* Add the binary RR_NS to the end of the udpzone string
                (This makes authoritative checks easier) */
-            if(js_addbyte(udpzone,0) == JS_ERROR)
+            if(js_addbyte(udpzone,0) == JS_ERROR) {
+                js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+                js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                js_destroy(data); js_destroy(zone); js_dealloc(desc);
                 return JS_ERROR;
-            if(js_addbyte(udpzone,RR_NS) == JS_ERROR)
+                }
+            if(js_addbyte(udpzone,RR_NS) == JS_ERROR) {
+                js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+                js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                js_destroy(data); js_destroy(zone); js_dealloc(desc);
                 return JS_ERROR;
+                }
 
             /* Get the file name to open from the element's value */
+            js_destroy(filename);
             filename = mhash_get_js(csvs,zone);
             /* Open up the filename in question */
             if(js_open_read(filename,desc) == JS_ERROR) {
@@ -698,13 +804,22 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
 
                 /* Get the line */
                 result = js_buf_getline(desc,line);
-                if(result == JS_ERROR)
+                if(result == JS_ERROR) {
+                    js_destroy(udpzone); js_destroy(filename);
+                    js_destroy(line);
+                    js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                    js_destroy(data); js_destroy(zone); js_dealloc(desc);
                     return JS_ERROR;
+                    }
                 else if(result == -2) {
                     printf("%s%d%s",L_LINE_NUMBER,linenum + 1,L_TOO_LONG);
                     show_esc_stdout(filename);
                     printf("\n");
                     /* "line number ... is too long in file " */
+                    js_destroy(udpzone); js_destroy(filename);
+                    js_destroy(line);
+                    js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                    js_destroy(data); js_destroy(zone); js_dealloc(desc);
                     return JS_ERROR;
                     }
 
@@ -720,9 +835,16 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
                         continue;
                     /* Add the SOA in the special case of the zone
                        only having SOA and NS records */
-                    if(parse_csv1_line(soaline,query,data,&ttl) != RR_SOA)
+                    if(parse_csv1_line(soaline,query,data,&ttl) != RR_SOA) {
+
+                        js_destroy(udpzone); js_destroy(filename);
+                        js_destroy(line);
+                        js_destroy(pline); js_destroy(soaline);
+                        js_destroy(query);
+                        js_destroy(data); js_destroy(zone); js_dealloc(desc);
                         return JS_ERROR;
-                    add_rr_to_bighash(maintable,query,data,ttl,udpzone);
+                        }
+                    add_rr_to_bighash(maintable,query,data,ttl,udpzone,0);
                     break;
                     }
 
@@ -731,8 +853,13 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
                     break;
 
                 /* Process the % character and any \ sequences */
-                if(bs_process(line,pline,zone) == JS_ERROR)
+                if(bs_process(line,pline,zone) == JS_ERROR) {
+                    js_destroy(udpzone); js_destroy(filename);
+                    js_destroy(line);
+                    js_destroy(pline); js_destroy(soaline); js_destroy(query);
+                    js_destroy(data); js_destroy(zone); js_dealloc(desc);
                     return JS_ERROR;
+                    }
 
                 /* Parse the line and get the rr type of the query */
                 rr_type = parse_csv1_line(pline,query,data,&ttl);
@@ -776,8 +903,15 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
                        record knows what NS records to place in the NS
                        section of the answer) */
                     else if(rrnum == 1) {
-                        if(js_copy(pline,soaline) == JS_ERROR)
+                        if(js_copy(pline,soaline) == JS_ERROR) {
+                            js_destroy(udpzone); js_destroy(filename);
+                            js_destroy(line);
+                            js_destroy(pline); js_destroy(soaline);
+                            js_destroy(query);
+                            js_destroy(data); js_destroy(zone);
+                            js_dealloc(desc);
                             return JS_ERROR;
+                            }
                         }
                     /* Only one SOA allowed per zone */
                     else if(rrnum != 1 && rr_type == RR_SOA) {
@@ -794,9 +928,17 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
                     else if((rr_type != RR_NS || !js_issame(query,udpzone))
                             && in_ns) {
                         /* Add the SOA after the zone's NS records */
-                        if(parse_csv1_line(soaline,query,data,&ttl) != RR_SOA)
+                        if(parse_csv1_line(soaline,query,data,&ttl) !=
+                           RR_SOA) {
+                            js_destroy(udpzone); js_destroy(filename);
+                            js_destroy(line);
+                            js_destroy(pline); js_destroy(soaline);
+                            js_destroy(query);
+                            js_destroy(data); js_destroy(zone);
+                            js_dealloc(desc);
                             return JS_ERROR;
-                        add_rr_to_bighash(maintable,query,data,ttl,udpzone);
+                            }
+                        add_rr_to_bighash(maintable,query,data,ttl,udpzone,0);
                         /* Now, add the record we are looking at (the
                            first non-SOA and non-authoritative-NS record) */
                         rr_type = parse_csv1_line(pline,query,data,&ttl);
@@ -817,21 +959,32 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
                             warning(error,linenum);
                             continue;
                             }
-                        add_rr_to_bighash(maintable,query,data,ttl,udpzone);
+                        add_rr_to_bighash(maintable,query,data,ttl,udpzone,0);
                         in_ns = 0;
                         }
                     else
-                        add_rr_to_bighash(maintable,query,data,ttl,udpzone);
+                        add_rr_to_bighash(maintable,query,data,ttl,udpzone,0);
                     }
                 }
             js_close(desc);
             }
         } while(mhash_nextkey(csvs,zone) != 0);
+    js_destroy(udpzone); js_destroy(filename); js_destroy(line);
+    js_destroy(pline); js_destroy(soaline); js_destroy(query);
+    js_destroy(data); js_destroy(zone); js_dealloc(desc);
+    return JS_SUCCESS;
+}
 
-    /* Pass 2: add the appropriate A records */
-    return add_an(maintable,error);
+/* This program parses all of the zone files and adds the records
+ * to the MaraDNS database */
+int populate_main(mhash *maintable, js_string *error, int recursive) {
+        /* Pass 1: parse all of the zone files */
+        parse_csv1s(maintable,error,recursive);
+        csv2_parse_main_bighash(maintable,0);
 
-    }
+        /* Pass 2: add the appropriate A records */
+        return add_an(maintable,error);
+}
 
 /* Warn about the existance of a DDIP record
    Input: pointer to query with ddip value
@@ -839,12 +992,33 @@ int populate_main(mhash *maintable, js_string *error, int recursive) {
 */
 
 int warn_ddip(js_string *query) {
-    if(log_level == 0)
+    int a = 0, l = 0;
+    if(log_level == 0 || query == 0 || query->string == 0)
         return JS_SUCCESS;
     /* "Dotted decimal IP for NS, CNAME, or MX does not work with some DNS servers" */
     printf("%s%s",L_DDIP_WARN,L_F);
-    return JS_SUCCESS;
+    printf("Hostname of record with problem: ");
+    for(a = 0 ; a < query->unit_count ; a++) {
+        l = *(query->string + a);
+        if(l < 1 || l > 64) {
+            printf("\n");
+            return JS_SUCCESS;
+        }
+        for(;l>0;l--) {
+            char c;
+            a++;
+            c = *(query->string + a);
+            if(c>' ' && c < '~') {
+                printf("%c",c);
+            } else {
+                printf("~");
+            }
+        }
+        printf(".");
     }
+    printf("\n");
+    return JS_SUCCESS;
+}
 
 /* Synthesize a DDIP record, just in case a MX, NS, or CNAME points to a
    dotted-decimal IP
@@ -904,12 +1078,12 @@ int make_ddip(mhash *bighash, js_string *query) {
     if((js_ip = js_create(5,1)) == 0)
         return JS_ERROR;
 
-    if(js_str2js(js_ip,ip,4,1) == JS_ERROR) {
+    if(js_str2js(js_ip,(char *)ip,4,1) == JS_ERROR) {
         js_destroy(js_ip);
         return JS_ERROR;
         }
 
-    if (mhash_put_rr(bighash,query,js_ip,8675309,1,0) == JS_ERROR) {
+    if(mhash_put_rr(bighash,query,js_ip,8675309,1,0,0) == JS_ERROR) {
         js_destroy(js_ip);
         return JS_ERROR;
         }
@@ -924,13 +1098,15 @@ int make_ddip(mhash *bighash, js_string *query) {
    AR section (AN section with CNAMEs)
    Input: A pointer to the hash to populate, a pointer to the string to
           put an error message in
-   Output: JS_ERROR on error, -2 on non-fatal error, JS_SUCCESS on success
+   Output: JS_ERROR on error, -2 on non-fatal error, JS_SUCCESS when there
+           is one or more authoritative records, 0 when there are no
+           authoritative records
  */
 
 int add_an(mhash *bighash, js_string *error) {
 
     js_string *query, *a_query;
-    mhash_e qdata, adata;
+    mhash_e qdata = {0,0,0}, adata = {0,0,0};
     rr *record;
     int qtype;
 
@@ -954,10 +1130,10 @@ int add_an(mhash *bighash, js_string *error) {
         qtype = get_rtype(query);
         if(qtype == JS_ERROR)
             continue;
-        /* If it is a MX/NS/CNAME query, proceed to find the assosciated
-           A value (if it exists) */
-        if(qtype == RR_MX || qtype == RR_NS || qtype == RR_CNAME) {
-            qdata = mhash_get(bighash,query);
+        qdata = mhash_get(bighash,query);
+        /* If it is a MX/NS query, proceed to find the assosciated
+           A (IPV4 address) value (if it exists) */
+        if(qtype == RR_MX || qtype == RR_NS) {
             if(qdata.datatype == MARA_DNSRR) {
                 /* See if we have a corresponding A record for each of
                    the answers */
@@ -985,7 +1161,40 @@ int add_an(mhash *bighash, js_string *error) {
                         }
                     } while ((record = record->next) != 0);
                 }
-            }
+            /* If it is a CNAME query, proceed to find the assosciated
+             * ANY value */
+            } else if(qtype == RR_CNAME) {
+                /* Look for a corresponding ANY value (which needs to be
+                 * there) */
+                record = qdata.value;
+                do {
+                    if(answer_ip_query(qtype,record->data,a_query) == JS_ERROR)
+                        continue;
+                    if(record->rr_type != RR_CNAME)
+                        break;
+                    if(change_rtype(a_query,RR_ANY) == JS_ERROR)
+                        continue;
+                    adata = mhash_get(bighash,a_query);
+                    /* If so, point to it */
+                    if(adata.datatype == MARA_DNS_LIST) {
+                        if(record != 0 && adata.value != 0)
+                            record->list = adata.value;
+                        }
+                    else if(no_cname_warnings == 0 &&
+                               cname_warnings_shown < 10) {
+                        printf("Warning: The CNAME record ");
+                        human_readable_dns_query(query,1);
+                        printf(" is a dangling CNAME record.\n"
+"Please read the FAQ entry about FAQ entry about dangling CNAME records.\n");
+                        cname_warnings_shown++;
+                        }
+                    else if(no_cname_warnings == 0 &&
+                              cname_warnings_shown == 10) {
+                       printf("Warning: More dangling CNAME records exist.\n");
+                       cname_warnings_shown++;
+                       }
+                    } while ((record = record->next) != 0);
+                }
         } while(mhash_nextkey(bighash,query) != 0);
     js_destroy(a_query);
     js_destroy(query);
@@ -1069,12 +1278,18 @@ int answer_ip_query(int qtype, js_string *qu, js_string *ar) {
              data in the database.  If it is authoritative, and authoritative
              data is there, do nothing.
 
+          The "perms" (what set of IPs are allowed to view this record) of
+          the record in question; if a given perm bit is one, only the IP
+          corresponding to the perm bit in question is allowed to view this
+          record.  If it is zero, this is the default value: any IP on
+          the internet can see the record in question.
+
    Output: JS_ERROR on error, JS_SUCCESS on success
 */
 
 int add_rr_to_bighash(mhash *bighash,
                       js_string *query, js_string *data, uint32 ttl,
-                      js_string *zone) {
+                      js_string *zone, perm_t perms) {
 
     int label_len,has_authority = 0,qtype;
     mhash_e hash_spot;
@@ -1091,6 +1306,17 @@ int add_rr_to_bighash(mhash *bighash,
     if((qtype = get_rtype(query)) == JS_ERROR)
         return JS_ERROR;
 
+    /* We are not allowed to add RR_ANY data to the bighash */
+    if(qtype == RR_ANY)
+        return JS_ERROR;
+
+    /* Nor are we allowed to add "AXFR" nor "IXFR" data to the bighash */
+    if(qtype == 251 /* IXFR */ || qtype == 252 /* AXFR */)
+        return JS_ERROR;
+
+    /* Increment the number of DNS records we are serving */
+    dns_records_served++;
+
     /* See if the data is authoritative or not */
     /* Change the type so the two strings are the same */
     if(change_rtype(zone,qtype) == JS_ERROR) {
@@ -1137,7 +1363,8 @@ int add_rr_to_bighash(mhash *bighash,
        the first NS record for this zone if this is a non-NS record.  */
     if(hash_spot.value == 0) {
         /* Add the data */
-        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0) == JS_ERROR)
+        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0,perms)
+                        == JS_ERROR)
             return JS_ERROR;
         if(mhash_add_ns(bighash,query,zone) == JS_ERROR)
             return JS_ERROR;
@@ -1158,7 +1385,8 @@ int add_rr_to_bighash(mhash *bighash,
        the beginning in the case of NS records) */
     if((point->authoritative == 1 && has_authority == 1) ||
        (point->authoritative == 0 && has_authority == 0)) {
-        if(mhash_add_rr(bighash,query,data,ttl,has_authority,0) == JS_ERROR)
+        if(mhash_add_rr(bighash,query,data,ttl,has_authority,0,perms)
+                        == JS_ERROR)
             return JS_ERROR;
         return JS_SUCCESS;
         }
@@ -1174,18 +1402,25 @@ int add_rr_to_bighash(mhash *bighash,
         nukeme = (rr *)mhash_undef(bighash,query);
         if(nukeme == 0)
             return JS_ERROR;
+        /* Remove this entry from the corresponding
+         * ANY chain */
+        if(any_zap_rr(bighash,query,nukeme) == JS_ERROR) {
+            return JS_ERROR;
+            }
+        /* Remove the entry once and for all */
         if(js_dealloc(nukeme) == JS_ERROR)
             return JS_ERROR;
 
-
         /* Add the data */
-        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0) == JS_ERROR)
+        if(mhash_put_rr(bighash,query,data,ttl,has_authority,0,perms)
+                        == JS_ERROR)
             return JS_ERROR;
         if(mhash_add_ns(bighash,query,zone) == JS_ERROR)
             return JS_ERROR;
         return JS_SUCCESS;
         }
 
+
     /* If the element exists, and it points to something authoritative, and
        we do not have authority for this record, then do nothing */
     if(point->authoritative == 1 && has_authority == 0)
